import numpy as np

from .PauliCode import PauliCode
from .QuantumCSSCode import QuantumCSSCode
from ...Math.BinaryArray import BinaryArray as ba
from ...Physics.MajoranaOperator import MajoranaOperator


class PauliCSSCode(PauliCode, QuantumCSSCode):

    # %%  USER：构造方法
    def __init__(self, generators_x, generators_z, physical_number):
        QuantumCSSCode.__init__(self, generators_x, generators_z, physical_number)

    # %%  USER：属性方法
    ##  TODO：求Pauli CSS code的距离（x方向）
    @property
    def distance_x(self):
        return 1

    ##  TODO：求Pauli CSS code的距离（z方向）
    @property
    def distance_z(self):
        return 1

    ##  TODO：求Pauli CSS code的逻辑算子（x方向）
    @property
    def logical_operators_x(self):
        return None

    ##  TODO：求Pauli CSS code的逻辑算子（z方向）
    @property
    def logical_operators_z(self):
        return None



import copy
import galois
import numpy as np
from mip import Model, xsum, minimize, BINARY


class BinaryArray:
    GF = galois.GF(2)

    #%%  USER：构造方法
    def __init__(self, array):
        if isinstance(array, BinaryArray):
            self._array = copy.deepcopy(array._array)  # 拷贝生成二进制数组
        else:
            self._array = self.GF(np.array(array, dtype=int))  # 转换为二进制数组
        self.shape = self._array.shape  # 数组形状

    #%%  USER：重载运算符
    ##  USER：返回数组的长度
    def __len__(self):
        return self.shape[0]

    ##  USER：判断两个数组是否相等
    def __eq__(self, other):
        if isinstance(other, BinaryArray):
            return self._array == other._array
        else:
            return self._array == other

    ##  USER：获取数组的元素
    def __getitem__(self, item):
        if isinstance(item, int):
            return self._array[item]
        else:
            return BinaryArray(self._array[item])

    ##  USER：设置数组的元素
    def __setitem__(self, item, value):
        self._array[item] = value

    ##  USER：返回数组的字符串表示
    def __str__(self):
        return str(self._array)

    ##  USER：数组相加
    def __add__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(self._array + other._array)

    ##  USER：数组相减
    def __sub__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(self._array - other._array)

    ##  USER：数组相乘
    def __mul__(self, other):
        assert (isinstance(other, BinaryArray)) or (isinstance(other, int))
        if isinstance(other, int):
            return BinaryArray(self._array * other)
        else:
            return BinaryArray(self._array * other._array)

    ##  USER：数组右乘
    def __rmul__(self, other):
        assert (isinstance(other, BinaryArray)) or (isinstance(other, int))
        if isinstance(other, int):
            return BinaryArray(self._array * other)
        else:
            return BinaryArray(self._array * other._array)

    ##  USER：数组矩阵乘
    def __matmul__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(self._array @ other._array)

    ##  USER：数组右矩阵乘
    def __rmatmul__(self, other):
        assert (isinstance(other, BinaryArray))
        return BinaryArray(other._array @ self._array)

    ##  USER：数组幂运算
    def __pow__(self, power):
        return BinaryArray(np.linalg.matrix_power(self._array, power))

    #%%  USER：属性方法
    ##  USER：返回数组的转置
    @property
    def T(self):
        return BinaryArray(self._array.T)

    ##  USER：求占据的位置
    @property
    def occupy(self):
        return np.where(self._array != 0)[0]

    ##  USER：返回数组的权重
    @property
    def weight(self):
        return np.count_nonzero(self._array)

    ##  USER：返回数组的零空间
    @property
    def null_space(self):
        return BinaryArray(self._array.null_space())

    ##  USER：返回数组的秩
    @property
    def rank(self):
        return np.linalg.matrix_rank(self._array)

    #%%  USER：对象方法
    ##  USER：复制函数
    def copy(self):
        return copy.deepcopy(self)

    #%%  USER：静态方法
    ##  USER：矩阵求和
    @staticmethod
    def sum(array_list):
        assert (isinstance(array_list, list))
        assert (isinstance(array_list[0], BinaryArray))
        return BinaryArray(np.sum([array._array for array in array_list], axis=0))

    ##  USER：求数组的水平堆叠
    @staticmethod
    def hstack(array, *args):
        assert len(args) > 0
        assert isinstance(array, BinaryArray)
        for temp in args:
            assert isinstance(temp, BinaryArray)
        result = array
        for temp in args:
            result = np.hstack((result, temp._array))
        return BinaryArray(result)

    ##  USER：求数组的垂直堆叠
    @staticmethod
    def vstack(array, *args):
        assert len(args) > 0
        assert isinstance(array, BinaryArray)
        for temp in args:
            assert isinstance(temp, BinaryArray)
        result = array
        for temp in args:
            result = np.vstack((result, temp._array))
        return BinaryArray(result)

    ##  USER：求解线性系统
    @staticmethod
    def solve(matrix, vector):
        assert isinstance(matrix, BinaryArray)
        assert isinstance(vector, BinaryArray)
        assert matrix.shape[1] == vector.shape[0]

        ##  使用GF(2)数组
        A = matrix._array.T
        b = vector._array.reshape(-1, 1)

        # 构造增广矩阵
        aug = np.concatenate((A, b), axis=1)
        n, m_plus_1 = aug.shape
        m = m_plus_1 - 1
        rank = 0

        ##  前向消元
        for col in range(m):
            ##  寻找主元
            pivot_row = None
            for i in range(rank, n):
                if aug[i, col] == 1:
                    pivot_row = i
                    break

            if pivot_row is None:
                ##  该列没有主元，跳过
                continue

            ##  交换行
            if pivot_row != rank:
                aug[[rank, pivot_row], :] = aug[[pivot_row, rank], :]

            ##  消去当前列下方和上方的元素
            for i in range(n):
                if i != rank and aug[i, col] == 1:
                    aug[i, :] ^= aug[rank, :]
            rank += 1

        ##  检查是否有解
        for i in range(rank, n):
            if aug[i, -1] == 1:
                ##  矛盾方程，无解
                return None

        ##  构造解向量
        solution = BinaryArray.GF.Zeros(m)
        leading_cols = []

        ##  找出主元列
        for i in range(rank):
            for j in range(m):
                if aug[i, j] == 1:
                    leading_cols.append(j)
                    break

        ##  回代求解
        for i in range(rank):
            col = leading_cols[i]
            solution[col] = aug[i, -1]

            ##  消去当前行中主元列右侧的元素对解的影响
            for j in range(col + 1, m):
                if aug[i, j] == 1:
                    solution[col] ^= solution[j]

        ##  返回解向量
        return BinaryArray(solution)

    ##  USER：返回两个线性空间的交集
    @staticmethod
    def cap(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)
        GF2 = galois.GF(2)

        ##  转化为GF2数组
        matrix1 = matrix1._array
        matrix2 = matrix2._array
        m = matrix1.shape[0]
        k = matrix2.shape[0]

        ##  解方程 basis1^T * x + basis2^T * y = 0 (表示交集向量), 构造矩阵 [basis1^T | basis2^T]
        aug_matrix = GF2(np.concatenate((matrix1.T, matrix2.T), axis=1))

        ##  计算零空间（解空间）
        nullspace = aug_matrix.null_space()

        ##  从零空间中取对应于每个解向量的系数（对应于方程组中的解向量）
        ab_space = nullspace[:, :m]

        ##  将系数乘以 basis1 得到具体解（交集）
        if len(ab_space) == 0:
            return GF2.Zeros((0, matrix1.shape[1]))

        intersection_vectors = ab_space @ matrix1

        ##  产生基向量的线性无关组合，即每种向量的不同的高斯行组合样本
        rref_intersection = intersection_vectors.row_reduce()

        ##  除去全0行的行（即为没有有用基底捕获的时候）
        nz_mask = np.any(rref_intersection != 0, axis=1)
        rref_basis = rref_intersection[nz_mask]

        ##  返回补集
        return BinaryArray(rref_basis)

    ##  USER：返回两个线性空间的差集
    @staticmethod
    def minus(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)

        ##  转化为GF2数组
        matrix1_origin = matrix1
        matrix2_origin = matrix2
        matrix1 = matrix1._array
        matrix2 = matrix2._array

        intersect = BinaryArray.cap(matrix1_origin, matrix2_origin)
        if len(intersect) == 0:
            return matrix1

        result = []
        for i in range(len(matrix1)):
            rank = np.linalg.matrix_rank(intersect)
            intersect = np.vstack((intersect, matrix1[i]))
            if np.linalg.matrix_rank(intersect) > rank:
                result.append(matrix1[i])

        ##  返回差集
        return BinaryArray(result)

    ##  USER：返回两个线性空间的直积
    @staticmethod
    def direct_sum(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)

        ##  转化为GF2数组
        matrix1 = matrix1._array
        matrix2 = matrix2._array
        result = matrix1[0].copy()

        ##  直接拼接
        for i in range(1, len(matrix1)):
            rank = np.linalg.matrix_rank(result)
            temp = np.vstack((result, matrix1[i]))
            if np.linalg.matrix_rank(temp) > rank:
                result = temp
        for i in range(len(matrix2)):
            rank = np.linalg.matrix_rank(result)
            temp = np.vstack((result, matrix2[i]))
            if np.linalg.matrix_rank(temp) > rank:
                result = temp

        ##  返回结果
        return BinaryArray(result)

    ##  USER：正交化基矢组
    @staticmethod
    def orthogonalize(matrix):
        def gf2_gram_schmidt(vectors, bilinear_form):
            k = len(vectors)
            ##  复制向量以避免修改原始数据
            B_i = [v.copy() for v in vectors]
            O = []  # 存储正交基
            for i in range(k):
                if np.mod(np.count_nonzero(B_i[i]), 2) == 1:
                    for j in range(k):
                        if j != i and np.mod(np.count_nonzero(B_i[j]), 2) == 0:
                            B_i[j] = B_i[j] + B_i[i]
                    break
                if i == k - 1:
                    print('There is no odd-weight vector')
                    return False

            while True:
                length = 10000
                b1 = None
                flag = 0
                b1 = B_i[0]
                # for i in range(len(B_i)):
                #     if length > np.count_nonzero(B_i[i]):
                #         b1 = B_i[i]
                #         flag = i
                #         length = np.count_nonzero(B_i[i])

                o_i = b1

                O.append(o_i)

                next_B = []
                for j in range(len(B_i)):
                    if j != flag:
                        b = B_i[j]
                        coef = bilinear_form(b, o_i)
                        b_new = b + coef * o_i
                        next_B.append(b_new)

                B_i = next_B
                if len(B_i) == 0:
                    return O
                length = len(B_i)
                for i in range(length):
                    if np.mod(np.count_nonzero(B_i[i]), 2) == 1:
                        for j in range(length):
                            if j != i and np.mod(np.count_nonzero(B_i[j]), 2) == 0:
                                B_i[j] = B_i[j] + B_i[i]
                        break
                    if i == len(B_i):
                        return O

        ##  双线性形式矩阵
        bilinear_form1 = lambda u, v: np.dot(u, v)
        ortho_basis1 = gf2_gram_schmidt(matrix, bilinear_form1)

        ##  输出结果
        return BinaryArray(ortho_basis1)

    ##  USER：计算张量积
    @staticmethod
    def kron(matrix1, matrix2):
        assert isinstance(matrix1, BinaryArray)
        assert isinstance(matrix2, BinaryArray)
        return BinaryArray(np.kron(matrix1._array, matrix2._array))

    ##  USER：计算code distance
    @staticmethod
    def distance(H, method):
        assert isinstance(H, BinaryArray)
        if method == 'mip':
            logic_op = BinaryArray.minus(H.null_space, H)
            return mip_distance_caculator(H._array, logic_op._array)
        elif method == 'random':
            return random_distance_caculator(H._array, H._array, 500, mindist=1, debug=1, field=2)

    ##  USER：计算子系统距离
    @staticmethod
    def subsystem_distance(stabilizers, gauges, method):
        pass

    ##  USER：根据占用位置生成二进制数组
    @staticmethod
    def FromOccupy(occupy, *args):
        if len(args) == 0:
            assert isinstance(occupy, list) or isinstance(occupy, np.ndarray)
            temp = np.zeros(np.max(occupy) + 1)
            temp[occupy] = 1
            return BinaryArray(temp)
        elif len(args) == 1:
            assert isinstance(occupy, list) or isinstance(occupy, np.ndarray)
            assert isinstance(args[0], int)
            temp = np.zeros(args[0], dtype=int)
            temp[occupy] = 1
            return BinaryArray(temp)
        else:
            raise ValueError("FromOccupy() takes 1 or 2 positional arguments but {} were given".format(len(args) + 1))

    ##  USER：根据数组生成二进制数组
    @staticmethod
    def FromArray(array):
        return BinaryArray(array)

    ##  USER：生成全零数组
    @staticmethod
    def zeros(number):
        return BinaryArray(np.zeros(number, dtype=int))

    ##  USER：生成单位矩阵
    @staticmethod
    def eye(number):
        return BinaryArray(np.eye(number, dtype=int))

    ##  USER：生成循环左移矩阵
    @staticmethod
    def shift(number, shift):
        S = np.zeros((number, number), dtype=int)
        for i in range(number):
            S[i, (i + shift) % number] = 1
        return BinaryArray(S)

    ##  USER：生成全一数组
    @staticmethod
    def ones(number):
        return BinaryArray(np.ones(number, dtype=int))


#%%  KEY：计算code distance
def mip_distance_caculator(H, logicOp):
    """"
    计算量子纠错码的代码距离

    该函数通过计算每个逻辑算子的最小权重，来确定量子纠错码的代码距离。
    代码距离是衡量量子码纠错能力的重要指标。

    Args:
        H: 稳定子生成器矩阵（二进制矩阵），表示量子码的稳定子群
        logicOp: 逻辑算子矩阵（二进制矩阵），每一行代表一个逻辑算子

    Returns:
        d: 量子码的代码距离，即所有逻辑算子中最小的非平凡权重
    """""
    ##  格式化输入
    H = np.array(H, dtype=int)  # 转换为整数类型的numpy数组
    logicOp = np.array(logicOp, dtype=int)  # 转换为整数类型的numpy数组
    d = H.shape[1]  # 初始化距离为量子比特数量（最大可能距离）

    ##  遍历每个逻辑算子
    for i in range(logicOp.shape[0]):
        logicOp_i = logicOp[i, :]
        n = H.shape[1]  # 量子比特数量（稳定子矩阵的列数）
        m = H.shape[0]  # 稳定子数量（稳定子矩阵的行数）
        wstab = np.max([np.sum(H[i, :]) for i in range(m)])  # 计算最大稳定子权重（单个稳定子中非零元素的最大数量）
        wlog = np.count_nonzero(logicOp_i)  # 计算逻辑算子的权重
        num_anc_stab = int(np.ceil(np.log2(wstab)))  # 计算稳定子约束所需的辅助变量数量（基于最大稳定子权重的对数）
        num_anc_logical = int(np.ceil(np.log2(wlog)))  # 计算逻辑算子约束所需的辅助变量数量（基于逻辑算子权重的对数）
        num_var = n + m * num_anc_stab + num_anc_logical  # 总变量数量 = 量子比特变量 + 稳定子辅助变量 + 逻辑算子辅助变量

        ##  创建混合整数规划模型
        model = Model()
        model.verbose = 0  # 关闭详细输出
        x = [model.add_var(var_type=BINARY) for i in range(num_var)]  # 创建二进制变量数组
        model.objective = minimize(xsum(x[i] for i in range(n)))  # 目标函数：最小化前n个变量（量子比特变量）的和（即最小化Hamming权重）

        # 为每个稳定子添加正交性约束（模2）
        for row in range(m):
            weight = [0] * num_var  # 初始化权重向量
            supp = np.nonzero(H[row, :])[0]  # 获取当前稳定子的支持集（非零元素的位置）

            ##  设置qubit变量的权重为1
            for q in supp:
                weight[q] = 1

            ##  添加辅助变量来处理模2约束
            cnt = 1
            for q in range(num_anc_stab):
                weight[n + row * num_anc_stab + q] = -(1 << cnt)  # 设置辅助变量的权重为负的2的幂次方
                cnt += 1
            model += xsum(weight[i] * x[i] for i in range(num_var)) == 0  # 添加约束：权重向量与变量向量的点积等于0

        ##  添加逻辑算子的奇数重叠约束
        supp = np.nonzero(logicOp_i)[0]  # 获取逻辑算子的支持集
        weight = [0] * num_var  # 初始化权重向量

        ##  设置qubit变量的权重为1
        for q in supp:
            weight[q] = 1

        ##  添加辅助变量来处理模2约束
        cnt = 1
        for q in range(num_anc_logical):
            # 设置辅助变量的权重为负的2的幂次方
            weight[n + m * num_anc_stab + q] = -(1 << cnt)
            cnt += 1

        ##  添加约束：权重向量与变量向量的点积等于1（奇数重叠）
        model += xsum(weight[i] * x[i] for i in range(num_var)) == 1

        ##  求解优化问题，计算最优解中前n个变量的和，即最小Hamming权重
        model.optimize()
        # noinspection PyTypeChecker
        opt_val = sum([x[i].x for i in range(n)])
        d = min(d, int(opt_val))

    ##  返回最小距离
    return d


#%%  KEY：计算随机距离
def random_distance_caculator(gx, gz, num, mindist=2, debug=0, field=None):
    """"
    使用随机算法计算量子纠错码的代码距离
    该函数通过随机排列和行约简技术来寻找满足特定约束的最小权重向量，
    从而估计量子码的代码距离。适用于CSS码（Calderbank-Shor-Steane codes）。

    Args:
        gx: X稳定子生成器矩阵（二进制矩阵）
        gz: Z稳定子生成器矩阵（二进制矩阵）
        num: 随机迭代次数，控制算法的精度
        mindist: 最小距离阈值，如果找到的距离小于等于此值则提前终止
        debug: 调试标志位，控制调试信息的输出
        field: 有限域，默认为GF(2)（二进制域）
        maxav: 是否记录最大权重向量的详细信息

    Returns:
        int: 估计的代码距离
    """""
    ##  设置默认有限域为GF(2)（二进制域）
    if field is None:
        F = galois.GF(2)
    else:
        F = galois.GF(field)

    ##  调试模式检查：验证X和Z稳定子是否正交（CSS码的基本要求）
    if debug & 2:
        if np.any(gx @ gz.T):
            raise ValueError("Input matrices must have orthogonal rows")

    wz = F(gx.null_space())  # 计算X稳定子的零空间（Z逻辑算子空间）
    wx = F(gz.null_space())  # 计算Z稳定子的零空间（X逻辑算子空间）
    rows_wz, cols_wz = wz.shape  # 获取零空间矩阵的维度信息
    dist_bound = cols_wz + 1  # 初始化距离上界为最大可能值（列数+1）
    vec_count = 0  # 计数器：记录找到当前最小权重的向量数量

    ##  主循环：进行num次随机迭代
    for i in range(num):
        per = np.random.permutation(cols_wz)  # 生成随机排列，用于随机化搜索顺序
        wz1 = wz[:, per]  # 对零空间矩阵的列进行随机排列
        wz2 = wz1.row_reduce()  # 对排列后的矩阵进行行约简（高斯消元）
        wz2 = wz2[:, np.argsort(per)]  # 将列顺序恢复为原始顺序

        ##  遍历行约简后的每一行
        for j in range(rows_wz):
            temp_vec = wz2[j, :]  # 获取当前行向量
            temp_weight = np.count_nonzero(temp_vec)  # 计算向量的Hamming权重（非零元素个数）

            ##  检查权重是否在有效范围内且小于等于当前最小距离
            if 0 < temp_weight <= dist_bound:

                ##  检查向量是否与X逻辑算子空间有非零重叠（即是否为非平凡逻辑算子）
                if np.any(wx @ temp_vec):

                    ##  如果找到更小的权重，更新最小距离
                    if temp_weight < dist_bound:
                        dist_bound = temp_weight
                        vec_count = 1

                    ##  如果权重等于当前最小距离，增加计数器
                    elif temp_weight == dist_bound:
                        vec_count += 1

            ##  检查是否达到最小距离阈值，如果达到则提前终止
            if dist_bound <= mindist:
                return mindist

    ##  返回找到的最小距离
    return dist_bound



from Circuit.Circuit import Circuit
from Code.QuantumCode.MajoranaCSSCode import MajoranaCSSCode


def MajoranaCSSCode2PhenomenologicalCircuit(code,p_noise,p_measure,cycle_number:int):
    assert isinstance(code,MajoranaCSSCode)
    stabilizers_x=code.generators_x
    stabilizers_z=code.generators_z
    logical_x=code.logical_operators_x
    logical_z=code.logical_operators_z
    logical_occupy=[]
    for i in range(len(logical_x)):
        logical_occupy.append(1j*logical_x[i]@logical_z[i])
    logical_string=[]
    for i in range(len(logical_x)):
        if i!=len(logical_x)-1:
            logical_string.append(logical_z[i]@logical_x[i+1])
        else:
            logical_string.append(logical_z[i+1]@logical_x[0])

    majorana_number=code.physical_number
    stabilizer_number = len(stabilizers_x) + len(stabilizers_z)
    circuit_z = Circuit()
    circuit_z.append("FR", range(majorana_number))

    ##  第一轮测量假设没有噪声，相当于完美的初始化
    for stabilizer in stabilizers_x:
        circuit_z.append("MPP", stabilizer)
    for stabilizer in stabilizers_z:
        circuit_z.append("MPP", stabilizer)

    ##  循环多轮，增加多轮测量错误与量子噪声信道
    for _ in range(cycle_number):
        for i in range(majorana_number):
            circuit_z.append("FDEPOLARIZE1", i, p_noise)

        for stabilizer in stabilizers_x:
            circuit_z.append("MPP", stabilizer,p_measure)
        for stabilizer in stabilizers_z:
            circuit_z.append("MPP", stabilizer,p_measure)

        for i in range(stabilizer_number):
            circuit_z.append("DETECTOR", [-i - 1, -i - stabilizer_number-1])

    ##  最后一轮测量假设是没有噪声的
    for stabilizer in stabilizers_x:
        circuit_z.append("MPP", stabilizer)
    for stabilizer in stabilizers_z:
        circuit_z.append("MPP", stabilizer)

    for i in range(stabilizer_number):
        circuit_z.append("DETECTOR", [-i - 1, -i - stabilizer_number - 1])

    ##  测量逻辑算符
    circuit_x=circuit_z.copy()
    circuit_z.append("MPP",logical_occupy)
    for i in range(len(logical_occupy)):
        circuit_z.append("OBSERVABLE_INCLUDE", [-i])
    circuit_x.append("MPP",logical_string)
    for i in range(len(logical_string)):
        circuit_x.append("OBSERVABLE_INCLUDE", [-i])
    return circuit_x, circuit_z





from abc import abstractmethod, ABC
import numpy as np
from Math.BinaryArray import BinaryArray as ba


class Operator(ABC):

    def __init__(self, occupy_x, occupy_z, coff):
        assert isinstance(occupy_x, np.ndarray) or isinstance(occupy_x, list)
        assert isinstance(occupy_z, np.ndarray) or isinstance(occupy_z, list)
        self.occupy_x = np.array(occupy_x, dtype=int)
        self.occupy_z = np.array(occupy_z, dtype=int)
        self.occupy_x.sort()
        self.occupy_z.sort()
        self.coff = coff

    #%%  USER：重载运算符
    ##  USER：矩阵乘法
    @abstractmethod
    def __matmul__(self, other):
        pass

    ##  USER：右矩阵乘法
    @abstractmethod
    def __rmatmul__(self, other):
        pass

    ##  USER：标量乘法
    @abstractmethod
    def __mul__(self, other):
        pass

    ##  USER：右标量乘法
    @abstractmethod
    def __rmul__(self, other):
        pass

    ##  USER：字符串表示
    @abstractmethod
    def __str__(self):
        pass

    ##  USER：相等判断
    @abstractmethod
    def __eq__(self, other):
        pass

    ##  USER：取负
    @abstractmethod
    def __neg__(self):
        pass

    #%%  USER：属性方法
    ##  USER：算符的权重
    @property
    def weight(self)->int:
        return len(self.occupy_x) + len(self.occupy_z)

    ##  USER：算符的类型
    @abstractmethod
    def type(self):
        return 'Operator'

    ##  USER：判断算符是否是厄米算符
    @property
    @abstractmethod
    def is_hermitian(self)->bool:
        pass

    ##  USER：求算符的对偶算符
    @property
    @abstractmethod
    def dual(self):
        pass

    #%%  USER：对象方法
    ##  USER：更换索引方式
    @abstractmethod
    def index_map(self, index):
        assert isinstance(index, np.ndarray) or isinstance(index, list)
        assert len(index) >= max(self.occupy_x.max(), self.occupy_z.max()) + 1
        x = np.array([index[i] for i in self.occupy_x], dtype=int)
        z = np.array([index[i] for i in self.occupy_z], dtype=int)
        return x, z, self.coff

    ##  USER：获取向量表示
    def get_vector(self, number):
        assert isinstance(number, int)
        vector = ba.zeros(number * 2)
        vector[self.occupy_x * 2] = 1
        vector[self.occupy_z * 2 + 1] = 1
        return vector

    ##  USER：检查是否存在量子位上的算符占据
    def is_exist_occupy_x(self, index):
        assert isinstance(index, int)
        occupy = np.where(self.occupy_x == index)[0]
        if len(occupy) == 0:
            return None
        return occupy[0]

    ##  USER：检查是否存在量子位上的算符占据
    def is_exist_occupy_z(self, index):
        assert isinstance(index, int)
        occupy = np.where(self.occupy_z == index)[0]
        if len(occupy) == 0:
            return None
        return occupy[0]

    ##  USER：复制算符
    @abstractmethod
    def copy(self):
        pass

    ##  USER：将算符在index位置切分
    @abstractmethod
    def split(self, index):
        assert isinstance(index, int)
        left_x=[temp for temp in self.occupy_x if temp < index]
        left_z=[temp for temp in self.occupy_z if temp < index]
        right_x=[temp for temp in self.occupy_x if temp > index]
        right_z=[temp for temp in self.occupy_z if temp > index]
        if self.is_exist_occupy_x(index) is not None and self.is_exist_occupy_z(index) is not None:
            middle_x=[index]
            middle_z=[index]
        elif self.is_exist_occupy_x(index) is not None:
            middle_x=[index]
            middle_z=[]
        elif self.is_exist_occupy_z(index) is not None:
            middle_x=[]
            middle_z=[index]
        else:
            middle_x=[]
            middle_z=[]
        return left_x, left_z, middle_x, middle_z, right_x, right_z

    #%%  USER：静态方法
    ##  USER：求多个算符的矩阵表示
    @staticmethod
    def get_matrix(ops, number):
        matrix = None
        for i, op in enumerate(ops):
            vector = op.get_vector(number)
            if i == 0:
                matrix = vector
            else:
                matrix = ba.vstack(matrix, vector)
        return matrix

    # %%  USER：静态方法
    ##  USER：定义一个厄米算符
    @staticmethod
    @abstractmethod
    def HermitianOperatorFromOccupy(occupy_x, occupy_z):
        pass

    ##  USER：定义一个厄米算符
    @staticmethod
    @abstractmethod
    def HermitianOperatorFromVector(vector):
        pass

    ##  USER：检查两个算符是否对易
    @staticmethod
    @abstractmethod
    def commute(A,B):
        pass



import copy
import qiskit
import numpy as np
import stim
import stimbposd
from qiskit.circuit import CircuitError
from qiskit.circuit.library import XGate
from Physics.MajoranaOperator import MajoranaOperator
from Physics.PauliOperator import PauliOperator
from Platform.Platform import Platform


class Circuit:
    #%%  USER：构造方法
    def __init__(self):
        self.majorana_number = 0
        self.pauli_number = 0
        self.sequence = []
        self.noise = []
        self.measurements = []
        self.detectors = []
        self.observables = []
        self._dem=None

    #%%  USER：重载运算符
    ##  USER：获取序列中的元素
    def __getitem__(self, item):
        return self.sequence[item]

    ##  USER：设置序列中的元素
    def __setitem__(self, key, value):
        self.sequence[key] = value

    #%%  USER：对象方法
    ##  USER：添加量子线路组分操作
    def append(self, name, target, *args):

        ##  添加single-qubit gate
        if name == 'X' or name == 'Y' or name == 'Z' or name == 'H' or name == 'S':
            assert len(args) == 0
            if isinstance(target, int):
                self.sequence.append({'name': name, 'target': target})
            else:
                for term in target:
                    self.sequence.append({'name': name, 'target': term})

        ##  添加single-fermionic-site gate
        elif name == 'U' or name == 'V' or name == 'N' or name == 'P':
            assert len(args) == 0
            if isinstance(target, int):
                self.sequence.append({'name': name, 'target': target})
            else:
                for term in target:
                    self.sequence.append({'name': name, 'target': term})

        ##  添加受控非门
        elif name == 'CX' or name == 'CUX' or name == 'CVX' or name == 'CNX':
            assert len(args) == 0
            assert len(target) == 2
            if isinstance(target[0],int) and len(target) == 2:
                self.sequence.append({'name': name, 'target': target})
            else:
                for i in range(len(target)):
                    self.sequence.append({'name': name, 'target': target[i]})

        ##  添加single-qubit and single-fermionic-site上的测量
        elif name == 'MZ' or name == 'MN':
            assert len(args) == 0 or len(args) == 1
            if name == 'MZ':
                self.sequence.append({'name': 'MPP', 'target': PauliOperator([], [target], 1)})
            else:
                self.sequence.append({'name': 'MPP', 'target': MajoranaOperator([target], [target], 1j)})
            self.measurements.append(len(self.sequence) - 1)
            if len(args) == 1:
                self.sequence.append({'name': 'M_ERROR', 'p': args[0]})
                self.noise.append(len(self.sequence) - 1)

        ##  添加qubit上的去极化噪声
        elif name == 'DEPOLARIZE1':
            assert len(args) == 1
            fix=27/26
            self.sequence.append({'name': 'X_ERROR', 'target': target, 'p': args[0]*fix / 3})
            self.noise.append(len(self.sequence) - 1)
            self.sequence.append({'name': 'Y_ERROR', 'target': target, 'p': args[0]*fix / 3})
            self.noise.append(len(self.sequence) - 1)
            self.sequence.append({'name': 'Z_ERROR', 'target': target, 'p': args[0]*fix / 3})
            self.noise.append(len(self.sequence) - 1)

        ##  添加fermionic-site上的去极化噪声
        elif name == 'FDEPOLARIZE1':
            assert len(args) == 1
            fix = 27 / 26
            self.sequence.append({'name': 'U_ERROR', 'target': target, 'p': args[0]*fix / 3})
            self.noise.append(len(self.sequence) - 1)
            self.sequence.append({'name': 'V_ERROR', 'target': target, 'p': args[0]*fix / 3})
            self.noise.append(len(self.sequence) - 1)
            self.sequence.append({'name': 'N_ERROR', 'target': target, 'p': args[0]*fix / 3})
            self.noise.append(len(self.sequence) - 1)

        ##  添加string算符的测量
        elif name == 'MPP':
            if isinstance(target, PauliOperator) or isinstance(target, MajoranaOperator):
                self.sequence.append({'name': name, 'target': target.copy()})
                self.measurements.append(len(self.sequence) - 1)
                if len(args) == 1:
                    self.sequence.append({'name': 'M_ERROR', 'p': args[0]})
                    self.noise.append(len(self.sequence) - 1)
            elif isinstance(target, list):
                for term in target:
                    self.sequence.append({'name': name, 'target': term.copy()})
                    self.measurements.append(len(self.sequence) - 1)
                    if len(args) == 1:
                        self.sequence.append({'name': 'M_ERROR', 'p': args[0]})
                        self.noise.append(len(self.sequence) - 1)

        ##  添加监视器
        elif name == 'DETECTOR':
            assert all(target[i] < 0 for i in range(len(target)))
            together = [len(self.measurements) + temp for temp in target]
            self.detectors.append(together)

        ##  添加可观测量
        elif name == 'OBSERVABLE_INCLUDE':
            assert all(target[i] < 0 for i in range(len(target)))
            together = [len(self.measurements) + temp for temp in target]
            self.observables.append(together)

        ##  添加qubit重置
        elif name == 'R':
            if isinstance(target, int):
                self.sequence.append({'name': name, 'target': target})
                if target == self.pauli_number:
                    self.pauli_number += 1
                elif target > self.pauli_number or target < 0:
                    raise ValueError("R gate target must be consecutive")
                else:
                    pass
            else:
                for term in target:
                    self.append('R',term)

        ##  添加fermionic-site重置
        elif name == 'FR':
            if isinstance(target, int):
                self.sequence.append({'name': name, 'target': target})
                if target == self.majorana_number :
                    self.majorana_number = target+1
                elif target > self.majorana_number or target < 0:
                    raise ValueError("FR gate target must be consecutive")
                else:
                    pass
            else:
                for term in target:
                    self.append('FR',term)
        else:
            raise NotImplementedError

    ##  USER：生成无噪声的线路
    def ideal_circuit(self):
        sequence = copy.deepcopy(self.sequence)
        for i in range(len(self.noise)):
            gate = sequence[self.noise[i]]
            assert isinstance(gate, dict)
            gate['p'] = 0
        ideal_circuit = Circuit()
        ideal_circuit.majorana_number = self.majorana_number
        ideal_circuit.pauli_number = self.pauli_number
        ideal_circuit.sequence = sequence
        ideal_circuit.measurements = self.measurements
        ideal_circuit.detectors = self.detectors
        ideal_circuit.observables = self.observables
        ideal_circuit.noise = self.noise
        return ideal_circuit

    ##  USER：执行线路并返回测量结果
    def execute(self):
        platform = Platform()
        platform.initialize(self.majorana_number, self.pauli_number)
        measurement_sample = np.empty(len(self.measurements), dtype=int)
        flag_measurement = 0
        for i, gate in enumerate(self.sequence):
            p = None
            target = None
            name = gate['name']
            if 'target' in gate:
                target = gate['target']
            if 'p' in gate:
                p = gate['p']
            if name == 'X':
                assert target is not None
                platform.X(target)
            elif name == 'Y':
                assert target is not None
                platform.Y(target)
            elif name == 'Z':
                assert target is not None
                platform.Z(target)
            elif name == 'H':
                assert target is not None
                platform.H(target)
            elif name == 'S':
                assert target is not None
                platform.S(target)
            elif name == 'U':
                assert target is not None
                platform.U(target)
            elif name == 'V':
                assert target is not None
                platform.V(target)
            elif name == 'N':
                assert target is not None
                platform.N(target)
            elif name == 'R':
                assert target is not None
                platform.reset(target)
            elif name == 'FR':
                assert target is not None
                platform.fermionic_reset(target)
            elif name == 'X_ERROR':
                assert target is not None
                assert p is not None
                platform.x_error(target, p)
            elif name == 'Y_ERROR':
                assert target is not None
                assert p is not None
                platform.y_error(target, p)
            elif name == 'Z_ERROR':
                assert target is not None
                assert p is not None
                platform.z_error(target, p)
            elif name == 'U_ERROR':
                assert target is not None
                assert p is not None
                platform.u_error(target, p)
            elif name == 'V_ERROR':
                assert target is not None
                assert p is not None
                platform.v_error(target, p)
            elif name == 'N_ERROR':
                assert target is not None
                assert p is not None
                platform.n_error(target, p)
            elif name == 'CX':
                assert target is not None
                platform.CX(target[0], target[1])
            elif name == 'CUX':
                assert target is not None
                platform.CUX(target[0], target[1])
            elif name == 'CVX':
                assert target is not None
                platform.CVX(target[0], target[1])
            elif name == 'CNX':
                assert target is not None
                platform.CNX(target[0], target[1])
            elif name == 'MPP':
                assert target is not None
                measurement_sample[flag_measurement] = platform.measure(target)
                flag_measurement += 1
            elif name == 'M_ERROR':
                assert p is not None
                if np.random.rand() < p:
                    measurement_sample[flag_measurement-1] = -measurement_sample[flag_measurement-1]
            else:
                raise NotImplementedError

        detector_sample = np.empty(len(self.detectors), dtype=bool)
        flag_detector = 0
        for i, detector in enumerate(self.detectors):
            value = measurement_sample[detector][0]
            detector_sample[flag_detector] = False
            for temp in measurement_sample[detector]:
                if value==temp:
                    continue
                else:
                    detector_sample[flag_detector]=True
                    break
            flag_detector += 1

        observable_sample = np.empty(len(self.observables), dtype=bool)
        flag_observable = 0
        for i, observable in enumerate(self.observables):
            if len(observable)==1:
                if measurement_sample[observable][0]==1:
                    observable_sample[flag_observable] = False
                else:
                    observable_sample[flag_observable] = True
            else:
                value = measurement_sample[observable][0]
                observable_sample[flag_observable] = False
                for temp in measurement_sample[observable]:
                    if value==temp:
                        continue
                    else:
                        observable_sample[flag_observable]=True
                        break
            flag_observable += 1
        return measurement_sample, detector_sample, observable_sample,platform

    ##  USER：生成检测错误模型
    def detector_error_model(self) -> stim.DetectorErrorModel:
        if self._dem is not None:
            return self._dem
        ideal_circuit = self.ideal_circuit()
        measurement_sample_origin, detector_sample_origin, observable_sample_origin,platform = ideal_circuit.execute()

        ##  执行检验线路的稳定性
        for time in range(5):
            measurement_sample, detector_sample, observable_sample,platform = ideal_circuit.execute()
            assert np.all(detector_sample==detector_sample_origin),f'原始线路的detector不是稳定的'
            assert np.all(observable_sample==observable_sample_origin),f'原始线路的observable不是稳定的'

        errors = []
        dem_str = ''
        for i in range(len(ideal_circuit.noise)):
            order = ideal_circuit.noise[i]
            gate_ideal = ideal_circuit.sequence[order]
            assert isinstance(gate_ideal, dict)
            gate_ideal['p'] = 1.1
            gate = self.sequence[order]
            assert isinstance(gate, dict)
            p = gate['p']
            measurement_sample, detector_sample, observable_sample,platform = ideal_circuit.execute()
            detector_sample_diff = [detector_sample_origin[j] ^ detector_sample[j] for j in range(len(detector_sample))]
            observable_sample_diff = [observable_sample_origin[j] ^ observable_sample[j] for j in range(len(observable_sample))]
            errors.append(len(errors))
            detectors_trigger = np.where(np.array(detector_sample_diff) == True)[0]
            observables_trigger = np.where(np.array(observable_sample_diff) == True)[0]
            if len(detectors_trigger) > 0 or len(observables_trigger) > 0:
                temp = f'error({p})'
                for index in detectors_trigger:
                    temp = temp + f' D{index}'
                for index in observables_trigger:
                    temp = temp + f' L{index}'
                dem_str += ('\n' + temp)
            gate_ideal['p'] = 0
        dem = stim.DetectorErrorModel(dem_str)
        self._dem=dem
        return dem

    ##  USER：生成编译后的采样器
    def compiler_sampler(self):
        dem = self.detector_error_model()
        return dem.compile_sampler()

    ##  USER：生成解码函数
    def decoder(self, method):
        dem = self.detector_error_model()
        if method == 'bposd':
            dec=stimbposd.bp_osd.BPOSD(model=dem,bp_method='minimum_sum')
        else:
            raise NotImplementedError
        return dec

    ##  USER：执行线路并返回错误率
    def sample(self,sample_number):
        sampler=self.compiler_sampler()
        decoder=self.decoder('bposd')
        detector_data, obs_data, error_data = sampler.sample(shots=sample_number)
        predictions = decoder.decode_batch(detector_data)
        num_errors = 0
        for shot in range(sample_number):
            actual_for_shot = obs_data[shot]
            predicted_for_shot = predictions[shot]
            if not np.array_equal(actual_for_shot, predicted_for_shot):
                num_errors += 1
        return num_errors/sample_number

    ##  USER：绘制线路图
    def draw(self, filename):
        # 绘制一个带有barriers和更多寄存器中，绘制一个新的电路
        F = qiskit.QuantumRegister(self.majorana_number, name='F')
        Q = qiskit.QuantumRegister(self.pauli_number, name='Q')
        C = qiskit.ClassicalRegister(1, name='C')
        A = qiskit.QuantumRegister(1, name='A')
        circuit_qiskit = qiskit.QuantumCircuit(F, Q, C,A)
        cux = qiskit.circuit.ControlledGate(name='CUX', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        cvx = qiskit.circuit.ControlledGate(name='CVX', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        cnx = qiskit.circuit.ControlledGate(name='CNX', num_qubits=2, params=[], label=None, num_ctrl_qubits=1, base_gate=XGate())
        x_error = qiskit.circuit.Gate('X', 1, label=None, params=[])
        y_error = qiskit.circuit.Gate('Y', 1, label=None, params=[])
        z_error = qiskit.circuit.Gate('Z', 1, label=None, params=[])
        u_error = qiskit.circuit.Gate('U', 1, label=None, params=[])
        v_error = qiskit.circuit.Gate('V', 1, label=None, params=[])
        n_error = qiskit.circuit.Gate('N', 1, label=None, params=[])
        m_error = qiskit.circuit.Gate('M', 1, label=None, params=[])
        rreset = qiskit.circuit.Gate('R', 1, label=None, params=[])
        for gate in self.sequence:
            if gate['name'] == 'R':
                circuit_qiskit.append(rreset, [Q[gate['target']]])
            elif gate['name'] == 'FR':
                circuit_qiskit.append(rreset, [F[gate['target']]])
            elif gate['name'] == 'X':
                circuit_qiskit.x(Q[gate['target']])
            elif gate['name'] == 'Y':
                circuit_qiskit.y(Q[gate['target']])
            elif gate['name'] == 'Z':
                circuit_qiskit.z(Q[gate['target']])
            elif gate['name'] == 'H':
                circuit_qiskit.h(Q[gate['target']])
            elif gate['name'] == 'S':
                circuit_qiskit.s(Q[gate['target']])
            elif gate['name'] == 'CX':
                circuit_qiskit.cx(Q[gate['target'][0]], Q[gate['target'][1]])
            elif gate['name'] == 'CUX':
                circuit_qiskit.append(cux, [F[gate['target'][0]],Q[gate['target'][1]]])
            elif gate['name'] == 'CVX':
                circuit_qiskit.append(cvx, [F[gate['target'][0]],Q[gate['target'][1]]])
            elif gate['name'] == 'CNX':
                circuit_qiskit.append(cnx, [F[gate['target'][0]],Q[gate['target'][1]]])
            elif gate['name'] == 'MPP':
                op = gate['target']
                if isinstance(op, MajoranaOperator):
                    f_flag_x = op.occupy_x
                    f_flag_z = op.occupy_z
                    f_flag_n = np.intersect1d(f_flag_x, f_flag_z)
                    f_flag_x = np.setdiff1d(f_flag_x, f_flag_n)
                    f_flag_z = np.setdiff1d(f_flag_z, f_flag_n)
                    f = np.concatenate([f_flag_x, f_flag_z, f_flag_n])
                    if len(f)>1:
                        mppx=qiskit.circuit.ControlledGate(name='MPPX', num_qubits=len(f)+1, params=[], label=None, num_ctrl_qubits=len(f), base_gate=XGate())
                        circuit_qiskit.append(mppx, F[f.tolist()]+[A[0]])
                        circuit_qiskit.measure(A[0], C[0])
                    else:
                        circuit_qiskit.measure(F[f[0]], C[0])
                elif isinstance(op, PauliOperator):
                    p_flag_x = op.occupy_x
                    p_flag_z = op.occupy_z
                    p_flag_y = np.intersect1d(p_flag_x, p_flag_z)
                    p_flag_x = np.setdiff1d(p_flag_x, p_flag_y)
                    p_flag_z = np.setdiff1d(p_flag_z, p_flag_y)
                    p = np.concatenate([p_flag_x, p_flag_y, p_flag_z])
                    if len(p)>1:
                        mppx=qiskit.circuit.ControlledGate(name='MPPX', num_qubits=len(p)+1, params=[], label=None, num_ctrl_qubits=len(p), base_gate=XGate())
                        circuit_qiskit.append(mppx, Q[p.tolist()]+[A[0]])
                        circuit_qiskit.measure(A[0], C[0])
                    else:
                        circuit_qiskit.measure(Q[p[0]], C[0])
                else:
                    raise CircuitError("cannot set parameters on immutable base gate")
            elif gate['name'] == 'M_ERROR':
                circuit_qiskit.append(m_error, [Q[gate['target']]])
            elif gate['name'] == 'X_ERROR':
                circuit_qiskit.append(x_error, [Q[gate['target']]])
            elif gate['name'] == 'Y_ERROR':
                circuit_qiskit.append(y_error, [Q[gate['target']]])
            elif gate['name'] == 'Z_ERROR':
                circuit_qiskit.append(z_error, [Q[gate['target']]])
            elif gate['name'] == 'U_ERROR':
                circuit_qiskit.append(u_error, [Q[gate['target']]])
            elif gate['name'] == 'V_ERROR':
                circuit_qiskit.append(v_error, [Q[gate['target']]])
            elif gate['name'] == 'N_ERROR':
                circuit_qiskit.append(n_error, [Q[gate['target']]])
        red='#E77081'
        blue='#5375CD'
        green='#00857B'
        grey='#8C92AC'
        purple='#5D548C'
        orange='#F15D22'
        pink='#FFACC5'
        cyan='#C9DCC4'
        circuit_qiskit.draw(output='mpl', filename=filename, style={
            'displaycolor': {'X': red, 'Y': red, 'Z': red,
                             'U': red, 'V': red, 'N': red,
                             'M': red, 'R': cyan,'measure':grey,
                             'x':blue, 'y':blue, 'z':blue,'s':blue,
                             'CNX':blue,'CUX':purple,'CVX':pink,
                             'cx':None,'MPPX':grey
                             },
            'fontsize': 15
        })

    ##  USER：复制函数
    def copy(self):
        return copy.deepcopy(self)



import QuantumCode
import QuantumCSSCode
import MajoranaCode
import MajoranaCSSCode
import PauliCode
import PauliCSSCode
import LatticeSurgery

QuantumCode=QuantumCode.QuantumCode
QuantumCSSCode=QuantumCSSCode.QuantumCSSCode
MajoranaCode=MajoranaCode.MajoranaCode
MajoranaCSSCode=MajoranaCSSCode.MajoranaCSSCode
PauliCode=PauliCode.PauliCode
PauliCSSCode=PauliCSSCode.PauliCSSCode



import numpy as np
from Math.BinaryArray import BinaryArray as ba
from Physics.MajoranaOperator import MajoranaOperator
from Physics.PauliOperator import PauliOperator


class Platform:

    # %%  USER：构造方法
    def __init__(self):
        self.pauli_number = 0
        self.majorana_number = 0
        self.stabilizers_pauli = []
        self.stabilizers_majorana = []

    # %%  USER：对象方法
    ##  USER：初始化平台，定义fermionic sites和qubits数目
    def initialize(self, majorana_number, pauli_number, *args):
        self.pauli_number = pauli_number
        self.majorana_number = majorana_number
        if len(args) == 0:
            for i in range(majorana_number):
                self.stabilizers_majorana.append(MajoranaOperator([i], [i], 1j))
                self.stabilizers_pauli.append(PauliOperator([], [], 1))
            for i in range(pauli_number):
                self.stabilizers_majorana.append(MajoranaOperator([], [], 1))
                self.stabilizers_pauli.append(PauliOperator([], [i], 1))
        elif len(args) == 2:
            self.stabilizers_majorana = args[0]
            self.stabilizers_pauli = args[1]

    ##  USER：测量算符op，返回测量结果
    def measure(self, op):
        if op.type == 'PauliOperator':
            stabilizers_now = self.stabilizers_pauli
        elif op.type == 'MajoranaOperator':
            stabilizers_now = self.stabilizers_majorana
        else:
            raise NotImplementedError
        assert op.is_hermitian

        first_pauli = None
        first_index = -1
        for i in range(len(stabilizers_now)):
            if op.type == 'PauliOperator':
                commute_flag = PauliOperator.commute(op, stabilizers_now[i])
            elif op.type == 'MajoranaOperator':
                commute_flag = MajoranaOperator.commute(op, stabilizers_now[i])
            else:
                raise NotImplementedError

            if not commute_flag:
                if first_index == -1:
                    first_pauli = stabilizers_now[i]
                    first_index = i
                else:
                    stabilizers_now[i] = stabilizers_now[i] @ first_pauli
        if first_index == -1:
            matrix_pauli = PauliOperator.get_matrix(self.stabilizers_pauli, self.pauli_number)
            matrix_majorana = MajoranaOperator.get_matrix(self.stabilizers_majorana, self.majorana_number)
            matrix = ba.hstack(matrix_majorana, matrix_pauli)
            if op.type == 'MajoranaOperator':
                vector_majorana = op.get_vector(self.majorana_number)
                vector_pauli = ba.zeros(self.pauli_number * 2)
                vector = ba.hstack(vector_majorana, vector_pauli)
            else:
                vector_majorana = ba.zeros(self.majorana_number * 2)
                vector_pauli = op.get_vector(self.pauli_number)
                vector = ba.hstack(vector_majorana, vector_pauli)
            result = ba.solve(matrix, vector)
            op_mul_pauli = PauliOperator([], [], 1)
            op_mul_majorana = MajoranaOperator([], [], 1)
            for i in range(len(result)):
                if result[i] == 1:
                    op_mul_majorana = op_mul_majorana @ self.stabilizers_majorana[i]
                    op_mul_pauli = op_mul_pauli @ self.stabilizers_pauli[i]
            coff = op_mul_pauli.coff * op_mul_majorana.coff
            if coff == op.coff:
                return 1
            else:
                return -1
        else:
            if np.random.rand() < 0.5:
                stabilizers_now[first_index] = op.copy()
                return 1
            else:
                stabilizers_now[first_index] = -op.copy()
                return -1

    ##  USER：X门，作用于qubit_index
    def X(self, qubit_index: int):
        for i in range(len(self.stabilizers_pauli)):
            if self.stabilizers_pauli[i].is_exist_occupy_z(qubit_index) is not None:
                self.stabilizers_pauli[i].coff = -self.stabilizers_pauli[i].coff

    ##  USER：Y门，作用于qubit_index
    def Y(self, qubit_index: int):
        for i in range(len(self.stabilizers_pauli)):
            if self.stabilizers_pauli[i].is_exist_occupy_x(qubit_index) is not None:
                self.stabilizers_pauli[i].coff = -self.stabilizers_pauli[i].coff
            if self.stabilizers_pauli[i].is_exist_occupy_z(qubit_index) is not None:
                self.stabilizers_pauli[i].coff = -self.stabilizers_pauli[i].coff

    ##  USER：Z门，作用于qubit_index
    def Z(self, qubit_index: int):
        for i in range(len(self.stabilizers_pauli)):
            if self.stabilizers_pauli[i].is_exist_occupy_x(qubit_index) is not None:
                self.stabilizers_pauli[i].coff = -self.stabilizers_pauli[i].coff

    ##  USER：Hadamard gate，作用于qubit_index
    def H(self, qubit_index: int):
        for i, stabilizer in enumerate(self.stabilizers_pauli):
            assert stabilizer.type == 'PauliOperator'
            left, middle, right = stabilizer.split(qubit_index)
            if len(middle.occupy_x) == 0 and len(middle.occupy_z) == 0:
                continue
            elif len(middle.occupy_x) == 1 and len(middle.occupy_z) == 1:
                continue
            elif len(middle.occupy_x) == 0 and len(middle.occupy_z) == 1:
                middle.occupy_x = [qubit_index]
                middle.occupy_z = []
            elif len(middle.occupy_x) == 1 and len(middle.occupy_z) == 0:
                middle.occupy_x = []
                middle.occupy_z = [qubit_index]
            else:
                raise ValueError
            self.stabilizers_pauli[i] = left @ middle @ right

    ##  USER：gamma门，作用于majorana_index
    def U(self, majorana_index: int):
        op = MajoranaOperator([majorana_index], [], 1)
        for i, stabilizer in enumerate(self.stabilizers_majorana):
            if not MajoranaOperator.commute(stabilizer, op):
                stabilizer.coff = -stabilizer.coff

    ##  USER：gamma_prime门，作用于majorana_index
    def V(self, majorana_index: int):
        op = MajoranaOperator([], [majorana_index], 1)
        for i, stabilizer in enumerate(self.stabilizers_majorana):
            if not MajoranaOperator.commute(stabilizer, op):
                stabilizer.coff = -stabilizer.coff

    ##  USER：i*gamma*gamma_prime门，作用于majorana_index
    def N(self, majorana_index: int):
        op = MajoranaOperator([majorana_index], [majorana_index], 1j)
        for i, stabilizer in enumerate(self.stabilizers_majorana):
            if not MajoranaOperator.commute(stabilizer, op):
                stabilizer.coff = -stabilizer.coff

    ##  USER：fermionic phase gate，作用于majorana_index
    def P(self, majorana_index: int):
        pass

    ##  USER：S门，作用于pauli_index
    def S(self, pauli_index: int):
        for i, stabilizer in enumerate(self.stabilizers_pauli):
            assert stabilizer.type=='PauliOperator'
            left, middle, right = stabilizer.split(pauli_index)
            if len(middle.occupy_x) == 0:
                continue
            elif len(middle.occupy_x) == 1 and len(middle.occupy_z) == 0:
                middle.occupy_z = [pauli_index]
                middle.coff = 1j
            elif len(middle.occupy_x) == 1 and len(middle.occupy_z) == 1:
                middle.occupy_z = []
                middle.coff = 1j
            else:
                raise ValueError
            self.stabilizers_pauli[i] = left @ middle @ right

    ##  USER：CNOT门，作用于control_index,target_index，两者是qubits，前者是控制位
    def CX(self, control_index, target_index):
        for i, stabilizer in enumerate(self.stabilizers_pauli):
            assert stabilizer.type == 'PauliOperator'
            left, middle, right = stabilizer.split(control_index)
            if len(middle.occupy_x) == 1:
                middle = PauliOperator([control_index, target_index], [], 1)
            if target_index > control_index:
                right_left, right_middle, right_right = right.split(target_index)
                if len(right_middle.occupy_z) == 1:
                    right_middle = PauliOperator([], [control_index, target_index], 1)
                self.stabilizers_pauli[i] = left @ middle @ right_left @ right_middle @ right_right
            elif target_index < control_index:
                left_left, left_middle, left_right = left.split(target_index)
                if len(left_middle.occupy_z) == 1:
                    left_middle = PauliOperator([], [control_index, target_index], 1)
                self.stabilizers_pauli[i] = left_left @ left_middle @ left_right @ middle @ right
            else:
                raise ValueError

    ##  USER：CN-NOT门，作用于control_index,target_index，前者是fermionic site控制位，后者是qubit目标位
    def CNX(self, control_index, target_index):
        for i in range(len(self.stabilizers_pauli)):
            stabilizer_pauli = self.stabilizers_pauli[i]
            stabilizer_majorana = self.stabilizers_majorana[i]
            left_control, middle_control, right_control = stabilizer_majorana.split(control_index)
            left_target, middle_target, right_target = stabilizer_pauli.split(target_index)
            majorana_product = PauliOperator([], [], 1)
            pauli_product = MajoranaOperator([], [], 1)
            if len(middle_control.occupy_x) == 1:
                majorana_product = PauliOperator([target_index], [], 1)
            if len(middle_control.occupy_z) == 1:
                majorana_product = majorana_product @ PauliOperator([target_index], [], 1)
            if len(middle_target.occupy_z) == 1:
                pauli_product = MajoranaOperator([target_index], [target_index], 1j)
            self.stabilizers_pauli[i] = majorana_product @ left_target @ middle_target @ right_target
            self.stabilizers_majorana[i] = left_control @ middle_control @ right_control @ pauli_product

    ##  USER：CU-NOT门，作用于control_index,target_index，前者是fermionic site控制位，后者是qubit目标位
    def CUX(self, control_index, target_index):
        for i in range(len(self.stabilizers_pauli)):
            stabilizer_pauli = self.stabilizers_pauli[i]
            stabilizer_majorana = self.stabilizers_majorana[i]
            left_control, middle_control, right_control = stabilizer_majorana.split(control_index)
            left_target, middle_target, right_target = stabilizer_pauli.split(target_index)
            majorana_product = PauliOperator([], [], 1)
            pauli_product = MajoranaOperator([], [], 1)
            if len(middle_control.occupy_z) == 1:
                majorana_product = PauliOperator([target_index], [], 1)
            if len(middle_target.occupy_z) == 1:
                pauli_product = MajoranaOperator([control_index], [], 1)
            self.stabilizers_pauli[i] = majorana_product @ left_target @ middle_target @ right_target
            self.stabilizers_majorana[i] = left_control @ middle_control @ right_control @ pauli_product

    ##  USER：CV-NOT门，作用于control_index,target_index，前者是fermionic site控制位，后者是qubit目标位
    def CVX(self, control_index, target_index):
        for i in range(len(self.stabilizers_pauli)):
            stabilizer_pauli = self.stabilizers_pauli[i]
            stabilizer_majorana = self.stabilizers_majorana[i]
            left_control, middle_control, right_control = stabilizer_majorana.split(control_index)
            left_target, middle_target, right_target = stabilizer_pauli.split(target_index)
            majorana_product = PauliOperator([], [], 1)
            pauli_product = MajoranaOperator([], [], 1)
            if len(middle_control.occupy_x) == 1:
                majorana_product = PauliOperator([target_index], [], 1)
            if len(middle_target.occupy_z) == 1:
                pauli_product = MajoranaOperator([], [control_index], 1)
            self.stabilizers_pauli[i] = majorana_product @ left_target @ middle_target @ right_target
            self.stabilizers_majorana[i] = left_control @ middle_control @ right_control @ pauli_product

    ##  USER：执行pauli_index上的X-error
    def x_error(self, pauli_index, p):
        if np.random.rand() < p:
            self.X(pauli_index)

    ##  USER：执行pauli_index上的Y-error
    def y_error(self, pauli_index, p):
        if np.random.rand() < p:
            self.Y(pauli_index)

    ##  USER：执行pauli_index上的Z-error
    def z_error(self, pauli_index, p):
        if np.random.rand() < p:
            self.Z(pauli_index)

    ##  USER：执行majorana_index上的U-error
    def u_error(self, majorana_index, p):
        if np.random.rand() < p:
            self.U(majorana_index)

    ##  USER：执行majorana_index上的V-error
    def v_error(self, majorana_index, p):
        if np.random.rand() < p:
            self.V(majorana_index)

    ##  USER：执行majorana_index上的N-error
    def n_error(self, majorana_index, p):
        if np.random.rand() < p:
            self.N(majorana_index)

    ##  USER：将系统在op上重置
    def clear(self, op):
        if isinstance(op, PauliOperator):
            stabilizers_now = self.stabilizers_pauli
        elif isinstance(op, MajoranaOperator):
            stabilizers_now = self.stabilizers_majorana
        else:
            raise NotImplementedError
        assert op.is_hermitian

        first_pauli = None
        first_index = -1
        for i in range(len(stabilizers_now)):
            if op.type == 'PauliOperator':
                commute_flag = PauliOperator.commute(op, stabilizers_now[i])
            elif op.type == 'MajoranaOperator':
                commute_flag = MajoranaOperator.commute(op, stabilizers_now[i])
            else:
                raise NotImplementedError

            if not commute_flag:
                if first_index == -1:
                    first_pauli = stabilizers_now[i]
                    first_index = i
                else:
                    stabilizers_now[i] = stabilizers_now[i] @ first_pauli
        if first_index == -1:
            matrix_pauli = PauliOperator.get_matrix(self.stabilizers_pauli, self.pauli_number)
            matrix_majorana = MajoranaOperator.get_matrix(self.stabilizers_majorana, self.majorana_number)
            if self.pauli_number == 0 and self.majorana_number != 0:
                matrix = matrix_majorana
            elif self.majorana_number == 0 and self.pauli_number != 0:
                matrix = matrix_pauli
            elif self.pauli_number != 0 and self.majorana_number != 0:
                matrix = ba.hstack(matrix_majorana, matrix_pauli)
            else:
                raise NotImplementedError

            if op.type=='MajoranaOperator':
                vector_majorana = op.get_vector(self.majorana_number)
                vector_pauli = ba.zeros(self.pauli_number * 2)
                vector = ba.hstack(vector_majorana, vector_pauli)
            else:
                vector_majorana = ba.zeros(self.majorana_number * 2)
                vector_pauli = op.get_vector(self.pauli_number)
                vector = ba.hstack(vector_majorana, vector_pauli)
            result = ba.solve(matrix, vector)
            op_mul_pauli = None
            op_mul_majorana = None
            flag = None
            for i in range(len(result)):
                if result[i] == 1:
                    if op_mul_pauli is None:
                        flag = i
                        op_mul_pauli = self.stabilizers_pauli[i]
                        op_mul_majorana = self.stabilizers_majorana[i]
                    else:
                        op_mul_pauli = op_mul_pauli @ self.stabilizers_pauli[i]
                        op_mul_majorana = op_mul_majorana @ self.stabilizers_majorana[i]
            if op_mul_pauli.coff * op_mul_majorana.coff == op.coff:
                pass
            else:
                assert flag is not None
                self.stabilizers_pauli[flag].coff = -self.stabilizers_pauli[flag].coff
        else:
            stabilizers_now[first_index] = op

    ##  USER：将系统在pauli_index上重置为0态
    def reset(self, pauli_index):
        op = PauliOperator([], [pauli_index], 1)
        self.clear(op)

    ##  USER：将系统在majorana_index上重置为空态
    def fermionic_reset(self, majorana_index):
        op = MajoranaOperator([majorana_index], [majorana_index], 1j)
        self.clear(op)


if __name__ == '__main__':
    from qutip import *
    gamma_0 = fcreate(2, 0) + fdestroy(2, 0)
    gamma_prime_0 = 1j * fdestroy(2, 0) - 1j * fcreate(2, 0)
    gamma_1 = fcreate(2, 1) + fdestroy(2, 1)
    gamma_prime_1 = 1j * fdestroy(2, 1) - 1j * fcreate(2, 1)

    T = (1j * (np.pi/2) * fcreate(2, 0)@fdestroy(2,1 ) +1j * (np.pi/2)*fcreate(2, 1) @fdestroy(2,0)).expm()

    all_list = [gamma_0 @ gamma_0,
                gamma_0, gamma_prime_0, gamma_1, gamma_prime_1,
                1j*gamma_0 @ gamma_prime_0, 1j*gamma_0 @ gamma_1, 1j*gamma_0 @gamma_prime_1, 1j*gamma_prime_0 @ gamma_1, 1j*gamma_prime_0@gamma_prime_1,
                1j*gamma_1 @ gamma_prime_1,
                1j*gamma_0 @ gamma_prime_0 @ gamma_1, 1j*gamma_0 @ gamma_prime_0 @ gamma_prime_1, 1*gamma_0 @ gamma_1 @ gamma_prime_1,
                1j*gamma_prime_0 @ gamma_1 @ gamma_prime_1,
                gamma_0 @ gamma_prime_0 @ gamma_1 @ gamma_prime_1,
                ]
    name = ['I',
            'gamma_0', 'gamma_prime_0', 'gamma_1', 'gamma_prime_1',
            'gamma_0@gamma_prime_0', 'gamma_0@gamma_1', 'gamma_0@gamma_prime_1','gamma_prime_0@gamma_1', 'gamma_prime_0@gamma_prime_1', 'gamma_1@gamma_prime_1',
            'gamma_0@gamma_prime_0@gamma_1','gamma_0@gamma_prime_0@gamma_prime_1', 'gamma_0@gamma_1@gamma_prime_1', 'gamma_prime_0@gamma_1@gamma_prime_1',
            'gamma_0@gamma_prime_0@gamma_1@gamma_prime_1']
    target = T @ gamma_prime_0 @ T.dag()
    for ii in range(len(all_list)):
        if (all_list[ii] - target).norm() < 1e-5:
            print(name[ii], 1)
        elif (-all_list[ii] - target).norm() < 1e-5:
            print(name[ii], -1)



import numpy as np
from Code.LinearCode.LinearCode import LinearCode
from Code.QuantumCode.MajoranaCode import MajoranaCode
from Code.QuantumCode.QuantumCSSCode import QuantumCSSCode
from Math.BinaryArray import BinaryArray as ba
from Physics.MajoranaOperator import MajoranaOperator


class MajoranaCSSCode(MajoranaCode, QuantumCSSCode):

    # %%  USER：构造方法
    def __init__(self, generators_x, generators_z, physical_number):
        QuantumCSSCode.__init__(self, generators_x, generators_z, physical_number)

    # %%  USER：属性方法
    ##  USER：求码距（x方向）
    @property
    def distance_x(self):
        return ba.distance(self.check_matrix_x,'mip')

    ##  USER：求码距（z方向）
    @property
    def distance_z(self):
        return ba.distance(self.check_matrix_z,'mip')

    ##  USER：求逻辑算符（x方向）
    @property
    def logical_operators_x(self):
        matrix = self.check_matrix_x()
        codewords = matrix.null_space()
        independent_null_basis_list = []
        for vec in codewords:
            rank_before = matrix.rank()
            matrix = ba.vstack(matrix, vec)
            if matrix.rank == rank_before + 1:
                independent_null_basis_list.append(vec)
        basis_list = ba.orthogonalize(independent_null_basis_list)
        majorana_logical_operators = []
        for i in range(len(basis_list)):
            occupy_x=np.where(basis_list[i]==1)[0]
            temp = MajoranaOperator.HermitianOperatorFromOccupy(occupy_x,[])
            majorana_logical_operators.append(temp)
        majorana_logical_operators = np.array(majorana_logical_operators, dtype=MajoranaOperator)
        return majorana_logical_operators

    ##  USER：求逻辑算符（z方向）
    @property
    def logical_operators_z(self):
        matrix = self.check_matrix_z()
        codewords = matrix.null_space()
        independent_null_basis_list = []
        for vec in codewords:
            rank_before = matrix.rank()
            matrix = ba.vstack(matrix, vec)
            if matrix.rank == rank_before + 1:
                independent_null_basis_list.append(vec)
        basis_list = ba.orthogonalize(independent_null_basis_list)
        majorana_logical_operators = []
        for i in range(len(basis_list)):
            occupy_z = np.where(basis_list[i] == 1)[0]
            temp = MajoranaOperator.HermitianOperatorFromOccupy([], occupy_z)
            majorana_logical_operators.append(temp)
        majorana_logical_operators = np.array(majorana_logical_operators, dtype=MajoranaOperator)
        return majorana_logical_operators

    #%%  USER：静态方法
    @staticmethod
    def FromCheckMatrix(check_matrix):
        generators_x = []
        generators_z = []
        for i in range(len(check_matrix)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromVector(check_matrix[i][0::2]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromVector(check_matrix[i][1::2]))
        physical_number=check_matrix.shape[1]
        return MajoranaCSSCode(generators_x, generators_z, physical_number)

    @staticmethod
    def FromLinearCode(linear_code):
        assert isinstance(linear_code,LinearCode)
        generators_x = []
        generators_z = []
        check_matrix=linear_code.check_matrix
        for i in range(len(check_matrix)):
            generators_x.append(MajoranaOperator.HermitianOperatorFromVector(check_matrix[i][0::2]))
            generators_z.append(MajoranaOperator.HermitianOperatorFromVector(check_matrix[i][1::2]))
        physical_number=check_matrix.shape[1]
        return MajoranaCSSCode(generators_x, generators_z, physical_number)


import LinearCode
import BicycleCode
import FiniteProjectiveGeometryCode
import FiniteEuclideanGeometryCode

LinearCode=LinearCode.LinearCode
BicycleCode=BicycleCode.BicycleCode
FiniteProjectiveGeometryCode=FiniteProjectiveGeometryCode.FiniteProjectiveGeometryCode
FiniteEuclideanGeometryCode=FiniteEuclideanGeometryCode.FiniteEuclideanGeometryCode



import numpy as np
from Math.BinaryArray import BinaryArray


class LinearCode:
    #%%  USER：构造方法
    def __init__(self,check_matrix):
        self.check_matrix = BinaryArray(check_matrix)
        self.number_bit=len(check_matrix[0])
        self.number_checker=len(check_matrix)

    #%%  USER：属性方法
    ##  USER：计算秩
    @property
    def rank(self):
        return self.check_matrix.rank

    ##  TODO：计算距离
    @property
    def distance(self):
        return 1

    ##  USER：计算逻辑位数目
    @property
    def logical_number(self):
        return self.number_bit-self.rank

    ##  USER：计算码字
    @property
    def codewords(self):
        return self.check_matrix.null_space

    ##  USER：计算dual linear code
    @property
    def dual(self):
        return LinearCode(self.check_matrix.null_space)

    ##  USER：判断是否dual-containing
    @property
    def is_dual_containing(self):
        return np.all(self.check_matrix @ self.check_matrix.T == 0)
